﻿//********************************************************************************
//***【NTL_WebSocketServer】Version 0.03  Copyright by Tomoaki Ueda (2014)    
//***
//***【注意事項】
//*** (1)本ソフトウェアは無償で私用または商用のいずれでもご利用になれます。
//*** (2)本ソフトウェアの改良行為はしていただいても構いません。
//*** (3)本ソフトウェアの再配布行為は許可いたしません。
//*** (4)本ソフトウェアの利用に伴う全ての結果やリスクに対して作者は責を負いません。
//***  また、技術サポートは行いません。利用者が全責任を負って使用される必要があり
//***  ます。
//********************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Threading.Tasks;
using System.Net;
using System.Net.WebSockets;

namespace WebSocketServer
{
    public partial class Form1 : Form
    {
        public static List<WebSocket> _client = new List<WebSocket>(); //【WebSocketで接続中のクライアントのリスト】
        public static string HostUri = "localhost:8080/";              //【ホストアドレス初期値】
        public static int receiveBufferSize = 0x10000;                 // バッファサイズは最大65536(0x10000)バイト
        private const int ImageWidth = 512;  //【転送しうる画像の水平方向画素数の最大値】
        private const int ImageHeight = 424; //【転送しうる画像の垂直方向画素数の最大値】
        public static byte[] ImageBuffer = new byte[ImageWidth * ImageHeight * 4]; //【送信バッファ】
        private Boolean fServer = false; //【サーバーの状態制御フラグ】

        public Form1()
        {
            InitializeComponent();

            textBox1.Text = HostUri;          //【ホストアドレス初期値をテキストボックスに読み込む】
            FormClosing += Form1_FormClosing; //【サーバー終了時の処理イベント】
        }

        private void button1_Click(object sender, EventArgs e)
        {
            if (!fServer)
            {   //【WebSocket Serverを起動する】

                button1.Text = "WebSocket強制終了";
                HostUri = textBox1.Text; //【指定ホストアドレスを読み込む】

                //【WebSocketサーバーのThreadを開始する】
                Thread WebSocketServerA = new Thread(new ThreadStart(WebSocketServer));
                WebSocketServerA.Start();

            }
            else
            {   //【WebSocket Serverを終了する】

                button1.Text = "WebSocket利用開始";

                AllSessionClose(); //【全てのセッションを閉じる】
                this.Close();
            }
            fServer = !fServer;
        }

        void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            AllSessionClose(); //【全てのセッションを閉じる】
        }

        //****************************************************************
        //****************************************************************
        //【WebSocketサーバー】
        //****************************************************************
        //****************************************************************

        //****************************************************************
        //【WebSocketサーバー】スタート　★この処理はasyncなので実行しつつ、次の処理も実行
        //****************************************************************
        public static async void WebSocketServer()
        {
            //【httpListenerでWebSocketによる接続を待機する】
            HttpListener httpListener = new HttpListener(); //【HttpListenerを取得】
            httpListener.Prefixes.Add("http://" + HostUri); //【localhost:8080で接続】
            httpListener.Start();                           //【HttpListener開始】

            while (true) //【WebSocket接続を待機するループ】
            {
                var listenerContext = await httpListener.GetContextAsync(); //【非同期操作で受信要求を待機】
                if (!listenerContext.Request.IsWebSocketRequest)
                { //【WebSocketでない場合はhttpレスポンスを返す】
                    listenerContext.Response.StatusCode = (int)HttpStatusCode.BadRequest; //【HTTPエラー400：要求の形式が正しくありません。】
                    listenerContext.Response.Close();                                     //【リソースを解放する】
                    return;
                }
                ProcessRequest(listenerContext); //【WebSocket接続毎の処理】
            }
        }

        //****************************************************************
        //【ProcessRequest】WebSocket接続毎の処理
        //****************************************************************
        public static async void ProcessRequest(HttpListenerContext listenerContext)
        {
            //【新規セッション開始】
            var wsc = await listenerContext.AcceptWebSocketAsync(
                subProtocol: null,
                receiveBufferSize: (int)receiveBufferSize,
                keepAliveInterval: new TimeSpan(100)
            ); //【ハンドシェーク成立を待機】
            var ws = wsc.WebSocket;                                                  //【WebSocket取得】
            _client.Add(ws); 　　　　　　　　　　　　　　　　　　　　　　　　　　　　//【新規クライアントをリストに追加】
            while (ws.State == WebSocketState.Open)
            { //【WebSocketの送受信ループ】
                try
                {
                    // ret.Count;             //パケット転送データバイト数
                    // ret.EndOfMessage=true; //パケット終端フラグ
                    int OffsetBytes = 0;
                    var ret = await ws.ReceiveAsync( //【最初のパケット受信を待機する】クライアント側の1パケットはデフォルトで16,384バイトのようだ。
                       new ArraySegment<byte>(ImageBuffer, OffsetBytes, (ImageBuffer.Length - OffsetBytes)),
                       System.Threading.CancellationToken.None
                    );
                    OffsetBytes += ret.Count;
                    if (!ret.EndOfMessage)
                    {
                        while (!ret.EndOfMessage) //【メッセージは終端パケットまで読み込んだか？】
                        {
                            ret = await ws.ReceiveAsync( //【最初のパケット受信を待機する】1パケットはデフォルトで16,384バイト
                               new ArraySegment<byte>(ImageBuffer, OffsetBytes, (ImageBuffer.Length - OffsetBytes)),
                               System.Threading.CancellationToken.None
                            );
                            OffsetBytes += ret.Count;
                        }
                    }
                    if (ret.MessageType == WebSocketMessageType.Binary)
                    { //【バイナリー】
                        Parallel.ForEach(_client, //【各クライアントへ配信】
                           p => p.SendAsync(
                              new ArraySegment<byte>(ImageBuffer, 0, OffsetBytes), //←【サイズ固定で良い】
                              WebSocketMessageType.Binary,
                              true,
                              System.Threading.CancellationToken.None
                           )
                        );
                    }
                    else if (ret.MessageType == WebSocketMessageType.Text)
                    { //【テキスト】
                        Parallel.ForEach(_client, //【各クライアントへ配信】
                           p => p.SendAsync(
                              new ArraySegment<byte>(ImageBuffer, 0, OffsetBytes),
                              WebSocketMessageType.Text,
                              true,
                              System.Threading.CancellationToken.None
                           )
                        );
                    }
                    else if (ret.MessageType == WebSocketMessageType.Close) { break; } //【セッション終了】
                }
                catch { break; } //【例外異常終了】
            }
            _client.Remove(ws); //【クライアント登録抹消】
            ws.Dispose();       //【WebSocketを破棄】
        }

        //****************************************************************
        //【AllSessionClose】WebSocket接続毎の処理
        //****************************************************************
        public void AllSessionClose()
        {
            Parallel.ForEach(_client, p => //【並列処理をすべて強制終了させる】
            {
                if (p.State == WebSocketState.Open)
                {
                    p.CloseAsync(WebSocketCloseStatus.NormalClosure, "", System.Threading.CancellationToken.None);
                }
            }
            );
        }

    }
}
