﻿//*************************************************************************************************
//【サンプルプログラム】『Kinectから骨格情報を取得するWebSocket Client』
//  Copyright by Tomoaki Ueda (/2014/05/06/)
//【注意事項】
// (1)本プログラムは
//　DepthImageの人物切り取り表示も行う
//【名称】KinectBasic_WebSocketClient　【Version】1.05
//【作成者】私設研究所Neo-Tech-Lab.com  上田智章
//【作成日】2014年5月6日
//【注意事項】
//　このプログラムをKinect for Windows SDK Version1.8で動かすには、参照設定の
// Microsoft.Kinectを一旦削除し、プロジェクト(P)-参照の追加(F)で、.NETタブ内
// にあるMicrosoft.Kinectを選択・追加してください。 
//*************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;                 //【Bitmap】【Color】
using System.Drawing.Imaging;         //【BitmapData】
using System.Drawing.Drawing2D;       //【2D系の描画】
using System.Diagnostics;             //【Debug】
using System.Runtime.InteropServices; //【Marshal.Copy】
using Microsoft.Kinect;               //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択
using System.Threading;               //【Thread】WebSocket Server
using System.Threading.Tasks;         //【Thread】WebSocket Server
using System.Net;                     //【WebSocket】
using System.Net.WebSockets;          //【WebSocket】
using NTL.KinectLib;                  //【デプス値の色変換】


namespace WebSocket_KinectBasic
{
    public partial class Form1 : Form
    {
        //【KINECT関連】
        //　関数内でダイナミックに変数を確保すると動作速度が遅くなるので、ここにまとめて変数を確保しています。
        private KinectSensor KS_KinectSensor;   //【KINECT】
        //【デプスイメージ関連】
        private DepthImageFrame KS_DepthFrame;  //【KINECTのDepthImageFrame】
        private short[] KS_DepthData;           //【KINECTのDepthImageData】16bit/画素(MSB13bit:Depth, LSB3bit:PlayerID) 
        public static byte[] KS_PixelDataARGB;  //【KINECTのDepthImageData】32bit/画素,A8R8G8B8表示形式
        public static int KS_DepthFrameWidth;   //【KINECTのDepthFrameの横方向画素数】
        public static int KS_DepthFrameHeight;  //【KINECTのDepthFrameの縦方向画素数】
        public static Boolean flagSend = false; //【WebSocket送信指示フラグ】
        public static Boolean fDepth = true;    //【WebSocketによるDepth転送開始フラグ】true:開始, false:終了
        private Bitmap KS_Bitmap;               //【DepthImage表示用Bitmapオブジェクト】
        private BitmapData KS_BitmapData;       //【DepthImage用BitmapData】
        private Rectangle KS_Rectangle;         //【DepthImage表示枠】
        //【デプス関連捕捉情報】
        private int KS_PlayerIndexBitmask;      //【DepthデータのPlayer情報】 0000 0000 0000 0111B = 7
        private int KS_PlayerIndexBitmaskWidth; //【DepthデータのPlayer情報】 3 bits
        private int KS_Lmin;                    //【深度測定可能最近距離】default:  800mm, near: 400mm 
        private int KS_Lmax;                    //【深度測定可能最遠距離】default: 4000mm, near:3000mm
        //【骨格トラッキング関連】
        private SkeletonFrame KS_SkeletonFrame; //【KINECTのSkeletonFrame】
        private Skeleton[] KS_SkeletonData;     //【KINECTのSkeleton】Skeleton 最大6名、20Joints/Player
        //【NTL拡張現実センサ関連】---------------------------------------------------------
        private NTL_ColorBar ColorBar = new NTL_ColorBar(); //【デプス値をカラーに変換する部品】
        private bool KS_flagFlip = false;       //【左右反転フラグ】
        private bool KS_flagPlayer = false;     //【Player検出フラグ】
        private SkeletonData NTL_SkeletonData;
        private int NTL_DepthCount = 0;
        
        public Form1()
        {
            InitializeComponent();
            this.Text = "デプス画像と骨格トラッキング";

            //***********************************************************
            //【Kinect検出】【共通】
            //***********************************************************
            //【利用可能なKINECTセンサを取得する】
            KS_KinectSensor = (from sensorToCheck
                               in KinectSensor.KinectSensors
                               where sensorToCheck.Status == KinectStatus.Connected
                               select sensorToCheck).FirstOrDefault();
            //【検出結果の表示】
            if (KS_KinectSensor == null) //【Kinect接続が確認できなかった場合】
            {
                pictureBox2.Image = Image.FromFile("../../image/RequiresKinectSensor.png");
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            pictureBox2.Image = Image.FromFile("../../image/KinectSensor.png");

            //***********************************************************
            //【Kinectの初期化処理】
            //***********************************************************
            //【KINECTの設定】DepthStreamとSkeletonStreamを許可***************************************************
         // KS_KinectSensor.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30); //【ColorStream許可】
            KS_KinectSensor.DepthStream.Enable(DepthImageFormat.Resolution320x240Fps30);    //【DepthStream許可】
            KS_KinectSensor.SkeletonStream.Enable();                                        //【SkeltonStream許可】
            //【★★★デプス情報のPlayerIDが有効になるのはSkeletonStreamを有効にしている場合に限られるので注意★★★】
            //【KINECT動作開始】**********************************************************************************
            KS_KinectSensor.Start(); //【KINECTスタート】

            //【KINECTデータ取得イベントの設定とデータバッファの確保】********************************************
            KS_KinectSensor.DepthFrameReady += new EventHandler<DepthImageFrameReadyEventArgs>(DepthImageReady); //【準備完了でDepthImageReadyがcallされる】
            KS_KinectSensor.AllFramesReady += new EventHandler<AllFramesReadyEventArgs>(KinectAllFramesReady);   //【準備完了でKinectAllFramesReadyがcallされる】

            //【デプスイメージ関連】
            KS_DepthData = new short[KS_KinectSensor.DepthStream.FramePixelDataLength];        //【深度情報データ用】
            KS_PixelDataARGB = new byte[KS_KinectSensor.DepthStream.FramePixelDataLength * 4]; //【深度情報データ表示用】
            KS_DepthFrameWidth = KS_KinectSensor.DepthStream.FrameWidth;   //【DepthImageの横方向画素数】
            KS_DepthFrameHeight = KS_KinectSensor.DepthStream.FrameHeight; //【DepthImageの縦方向画素数】
            KS_Bitmap = new Bitmap(KS_DepthFrameWidth, KS_DepthFrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【DepthImage表示用】
            KS_Rectangle = new Rectangle(0, 0, KS_DepthFrameWidth, KS_DepthFrameHeight); //【DepthImage用枠】

            //【骨格トラッキング関連】
            KS_SkeletonData = new Skeleton[KS_KinectSensor.SkeletonStream.FrameSkeletonArrayLength]; //【骨格情報用】
            KS_PlayerIndexBitmask = DepthImageFrame.PlayerIndexBitmask;
            KS_PlayerIndexBitmaskWidth = DepthImageFrame.PlayerIndexBitmaskWidth;
            KS_KinectSensor.DepthStream.Range = DepthRange.Default;    //【Depth測定モードをDefaultに設定】
            KS_CheckDistance();                                        //【設定結果を表示】Lmax, Lminを取得

            //【Kinectヘッドの仰角を設定】設定範囲は±27°
            int Angle = 0;
            KS_KinectSensor.ElevationAngle = Angle;
            textBox1.Text = Angle.ToString();

            //【表示関連】
            pictureBox1.SizeMode = PictureBoxSizeMode.StretchImage; //【デプス画像表示用pictureBox】
            pictureBox2.SizeMode = PictureBoxSizeMode.StretchImage; //【センサ検出結果用pictureBox】
            pictureBox3.SizeMode = PictureBoxSizeMode.StretchImage; //【カラースケール用pictureBox】
            ColorBar.CreateColorPalletForDistance();                       //【カラーパレットを作成する】
            pictureBox3.Image = ColorBar.CreateColorScale(320, 10, false); //【320×10画素サイズのカラーバーを作成・表示】

            NTL_SkeletonData = new SkeletonData();

            radioButton1.CheckedChanged += new EventHandler(DepthRangeChange);      //【深度測定モード変更イベント】
            checkBox1.CheckedChanged += new EventHandler(checkBox1_CheckedChanged); //【左右反転表示用イベント】
            this.FormClosing += new FormClosingEventHandler(Form1_FormClosing);     //

            //【WebSocketServerを起動する】
            Thread WebSocketServerA = new Thread(new ThreadStart(WebSocketServer));
            WebSocketServerA.Start(); //【注：WebSocketクライアント起動はこの後で行うこと】

            System.Threading.Thread.Sleep(3000);
            System.Diagnostics.Process.Start("http://neotechlab.jp/WebSocket/WebSocketClient.htm");
        }

        //***************************************************************
        //【上下角変更ボタン】KINECTの上下角を変更。±27°の範囲内。時間を置かないとエラーが発生する
        //***************************************************************
        private void button1_Click(object sender, EventArgs e)
        {
            Int32 angle = int.Parse(textBox1.Text); //【設定角度を読み込む】±27°
            if (angle > KS_KinectSensor.MaxElevationAngle || angle < KS_KinectSensor.MinElevationAngle)
            {   //【設定可能範囲外】
                angle = 0;
                textBox1.Text = angle.ToString();
            }
            KS_KinectSensor.ElevationAngle = angle; //【上下角の変更】
        }

        //***************************************************************
        //【WebSocket制御用ボタン】
        //***************************************************************
        private void button2_Click(object sender, EventArgs e)
        {
            if (fDepth)
            {
                WebSocketClient(); //【WebSocketクライアント起動】
                Thread ClientReceiveA = new Thread(new ThreadStart(ClientReceive));     //【受信待機スレッド準備】
                Thread ClientSendImageA = new Thread(new ThreadStart(ClientSendImage)); //【送信待機スレッド準備】
                ClientReceiveA.Start();   //【受信待機】
                ClientSendImageA.Start(); //【送信待機】
                button2.Text = "Kinectソケット切断";
                fDepth = false;
            }
            else
            {
                this.Close();
            }
        }

        //***************************************************************
        //【終了処理】
        //***************************************************************
        void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (KS_KinectSensor != null)
            {
                ClientSessionClose(); //【KinectのWebSocketClientセッションを強制終了する】

                KS_KinectSensor.Stop();    //【KINECTセンサを終了】★これを実行しないとレーザーが消えない
                KS_KinectSensor.Dispose(); //【KINECTセンサのオブジェクトを解放する】

                AllSessionClose(); //【全てのWebSocketのセッションを強制終了する】
            }
        }

        //***************************************************************
        //【距離モード変更】Defaultモード(800mm～4000mm) Nearモード(400m～3000mm)
        //***************************************************************
        void DepthRangeChange(object sender, EventArgs e)
        {
            if (radioButton1.Checked) { KS_KinectSensor.DepthStream.Range = DepthRange.Default; label3.Text = "Default"; }
            else { KS_KinectSensor.DepthStream.Range = DepthRange.Near; label3.Text = "Near"; }
            KS_CheckDistance();
        }

        //***************************************************************
        //【CheckDistance】KINECTセンサの検出可能距離現在値を調べる
        //***************************************************************
        private void KS_CheckDistance()
        {
            KS_Lmax = KS_KinectSensor.DepthStream.MaxDepth;
            KS_Lmin = KS_KinectSensor.DepthStream.MinDepth;
            label1.Text = "Min=" + KS_Lmin.ToString() + "mm";
            label2.Text = "Max=" + KS_Lmax.ToString() + "mm";
        }

        //***************************************************************
        //【カラー画像の左右反転指定フラグ】
        //***************************************************************
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }
        }

        //***************************************************************
        //【DepthImageの取得イベント】
        //***************************************************************
        private void DepthImageReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            using (KS_DepthFrame = e.OpenDepthImageFrame())
            {
                if (KS_DepthFrame != null)
                {
                    //kinectSensor.DepthStreamに以下のフラグがある。
                    // .TooNearDepth【距離が近すぎる】, .TooFarDepth【距離が遠すぎる】, .UnknownDepth【距離が不明】
                    KS_DepthFrame.CopyPixelDataTo(KS_DepthData); //【KINECTのDepth情報を配列DepthDataにコピー】
                    KS_ConvertDepthToColor(); //【深度情報をカラーバーの色を使って表示形式をA8R8G8B8形式に変換】

                    KS_BitmapData = KS_Bitmap.LockBits(KS_Rectangle, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【Lockする】
                    Marshal.Copy(KS_PixelDataARGB, 0, KS_BitmapData.Scan0, KS_PixelDataARGB.Length); //【UnmanagedMemoryのBitmapDataに配列depthDataをコピーする】
                    KS_Bitmap.UnlockBits(KS_BitmapData); //【Lock解除】
                    if (KS_flagFlip) { KS_Bitmap.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox1.Image = KS_Bitmap;  //【pictureBox1に画像表示】

                    if (NTL_DepthCount >= 10)
                    {
                        if (flagSend == false) { flagSend = true; }
                    }
                    NTL_DepthCount++;
                }
                else
                {
                    //【DepthFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        //***************************************************************
        //【ConvertDepthToColor】16ビットのDepth値をカラーバーに対応する色に変更
        //***************************************************************
        private void KS_ConvertDepthToColor()
        {
            int i, j, d;  //【Nearモード】  400～3000mm ⇒ 0.0980769　2600色
            int playerID; //【下位3bitはplayerID】
            int f; //【PlayerLost検出用】
            if (KS_flagPlayer == false)
            {
                j = 0;
                for (i = 0; i < KS_DepthData.Length; i++)
                {
                    d = KS_DepthData[i];                  //【距離情報】【上位13bitがDepth値】
                    playerID = d & KS_PlayerIndexBitmask; //【下位3bitはplayerID】【未使用】
                    if (playerID > 0) { KS_flagPlayer = true; }
                    d = ((d >> KS_PlayerIndexBitmaskWidth) & 0x1FFF);  //【realDepth】0mm～8191mm
                    if (d < KS_Lmin) { d = KS_Lmin; } //【近すぎる】
                    if (d > KS_Lmax) { d = KS_Lmax; } //【遠すぎる】
                    KS_PixelDataARGB[j    ] = ColorBar.B[d]; //【青】B
                    KS_PixelDataARGB[j + 1] = ColorBar.G[d]; //【緑】G
                    KS_PixelDataARGB[j + 2] = ColorBar.R[d]; //【赤】R
                    KS_PixelDataARGB[j + 3] = ColorBar.A[d]; //【α】
                    j = j + 4;
                }
            }
            else
            {
                f = 0;
                j = 0;
                for (i = 0; i < KS_DepthData.Length; i++)
                {
                    d = KS_DepthData[i];                  //【距離情報】【上位13bitがDepth値】
                    playerID = d & KS_PlayerIndexBitmask; //【下位3bitはplayerID】【未使用】
                    if (playerID > 0)
                    {
                        f = 1;
                        d = ((d >> KS_PlayerIndexBitmaskWidth) & 0x1FFF);  //【realDepth】0mm～8191mm
                        if (d < KS_Lmin) { d = KS_Lmin; } //【近すぎる】
                        if (d > KS_Lmax) { d = KS_Lmax; } //【遠すぎる】
                        KS_PixelDataARGB[j    ] = ColorBar.B[d]; //【青】B
                        KS_PixelDataARGB[j + 1] = ColorBar.G[d]; //【緑】G
                        KS_PixelDataARGB[j + 2] = ColorBar.R[d]; //【赤】R
                        KS_PixelDataARGB[j + 3] = ColorBar.A[d]; //【α】
                    }
                    else
                    {   //【背景画像を透過させるために透明にする】
                        KS_PixelDataARGB[j    ] = 0; //【青】B
                        KS_PixelDataARGB[j + 1] = 0; //【緑】G
                        KS_PixelDataARGB[j + 2] = 0; //【赤】R
                        KS_PixelDataARGB[j + 3] = 1; //【α】
                    }
                    j = j + 4;
                }
                if (f == 0) { KS_flagPlayer = false; }
            }
        }

        //***************************************************************
        //【SkeletonFrameの取得イベント】SkeletonFrameは6名分。各々がJointsコレクションを持つ。
        //***************************************************************
        private void KinectAllFramesReady(object sender, AllFramesReadyEventArgs e)
        {
            using (KS_SkeletonFrame = e.OpenSkeletonFrame())
            {
                if (KS_SkeletonFrame != null)
                {
                    KS_SkeletonFrame.CopySkeletonDataTo(KS_SkeletonData); //【KINECTのSkeltonデータを配列SkeletonDataにコピー】
                    for (int i = 0; i < KS_SkeletonData.Length; i++)
                    {
                        if (KS_SkeletonData[i].TrackingState == SkeletonTrackingState.Tracked)
                        {
                            label4.Text = "検出したプレイヤー番号＝" + i.ToString();
                            string s = NTL_SkeletonData.SetBody(KS_KinectSensor, KS_SkeletonData[i], KS_flagFlip);
                            break;
                        }
                    }
                }
                else
                {
                    //【SkeletonFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }





        //****************************************************************
        //****************************************************************
        //【WebSocketサーバー】
        //****************************************************************
        //****************************************************************
        public const int KS_FRAMEWIDTH = 320;  //【デプス画像水平方向画素数】
        public const int KS_FRAMEHEIGHT = 240; //【デプス画像垂直方向画素数】
        public static List<WebSocket> _client = new List<WebSocket>(); //【WebSocketで接続中のクライアントのリスト】
        public static string HostUri = "http://localhost:8080/";       //【ホストアドレス】
        public static int receiveBufferSize = 65536;
        public static byte[] ServerBuffer = new byte[KS_FRAMEWIDTH * KS_FRAMEHEIGHT * 4];    //【送信バッファ】320×240画素分のバッファを用意
        //   public static ArraySegment<byte> ImgBuffer = new ArraySegment<byte>(ImageBuffer);
        public static Boolean ClientFlag = false;

        //****************************************************************
        //【WebSocketサーバー】スタート　★この処理はasyncなので実行しつつ、次の処理も実行
        //****************************************************************
        public static async void WebSocketServer()
        {
            //【httpListenerでWebSocketによる接続を待機する】
            HttpListener httpListener = new HttpListener(); //【HttpListenerを取得】
            httpListener.Prefixes.Add(HostUri);             //【localhost:8080で接続】
            httpListener.Start();                           //【HttpListener開始】

            while (true) //【WebSocket接続を待機するループ】
            {
                var listenerContext = await httpListener.GetContextAsync(); //【非同期操作で受信要求を待機】
                if (!listenerContext.Request.IsWebSocketRequest)
                { //【WebSocketでない場合はhttpレスポンスを返す】
                    listenerContext.Response.StatusCode = (int)HttpStatusCode.BadRequest; //【HTTPエラー400：要求の形式が正しくありません。】
                    listenerContext.Response.Close();                                     //【リソースを解放する】
                    return;
                }
                ProcessRequest(listenerContext);
            }
        }

        //****************************************************************
        //【ProcessRequest】WebSocket接続毎の処理
        //****************************************************************
        public static async void ProcessRequest(HttpListenerContext listenerContext)
        {
            //【新規セッション開始】
            var wsc = await listenerContext.AcceptWebSocketAsync(
                subProtocol: null,
                receiveBufferSize: (int)receiveBufferSize,
                keepAliveInterval: new TimeSpan(100)
            ); //【ハンドシェーク成立を待機】
            var ws = wsc.WebSocket;                                                  //【WebSocket取得】
            _client.Add(ws); 　　　　　　　　　　　　　　　　　　　　　　　　　　　　//【新規クライアントをリストに追加】
            while (ws.State == WebSocketState.Open)
            { //【WebSocketの送受信ループ】
                try
                {
                    // ret.Count;             //パケット転送データバイト数
                    // ret.EndOfMessage=true; //パケット終端フラグ
                    int OffsetBytes = 0;
                    var ret = await ws.ReceiveAsync( //【最初のパケット受信を待機する】クライアント側の1パケットはデフォルトで16,384バイトのようだ。
                       new ArraySegment<byte>(ServerBuffer, OffsetBytes, (ServerBuffer.Length - OffsetBytes)),
                       System.Threading.CancellationToken.None
                    );
                    OffsetBytes += ret.Count;
                    if (!ret.EndOfMessage)
                    {
                        while (!ret.EndOfMessage) //【メッセージは終端パケットまで読み込んだか？】
                        {
                            ret = await ws.ReceiveAsync( //【最初のパケット受信を待機する】1パケットはデフォルトで16,384バイト
                               new ArraySegment<byte>(ServerBuffer, OffsetBytes, (ServerBuffer.Length - OffsetBytes)),
                               System.Threading.CancellationToken.None
                            );
                            OffsetBytes += ret.Count;
                        }
                    }
                    if (ret.MessageType == WebSocketMessageType.Binary)
                    { //【バイナリー】
                        Parallel.ForEach(_client, //【各クライアントへ配信】
                           p => p.SendAsync(
                              new ArraySegment<byte>(ServerBuffer, 0, OffsetBytes), //←【サイズ固定で良い】
                              WebSocketMessageType.Binary,
                              true,
                              System.Threading.CancellationToken.None
                           )
                        );
                    }
                    else if (ret.MessageType == WebSocketMessageType.Text)
                    { //【テキスト】
                        Parallel.ForEach(_client, //【各クライアントへ配信】
                           p => p.SendAsync(
                              new ArraySegment<byte>(ServerBuffer, 0, OffsetBytes),
                              WebSocketMessageType.Text,
                              true,
                              System.Threading.CancellationToken.None
                           )
                        );
                    }
                    else if (ret.MessageType == WebSocketMessageType.Close) { break; } //【セッション終了】
                }
                catch { break; } //【例外異常終了】
            }
            _client.Remove(ws); //【クライアント登録抹消】
            ws.Dispose();       //【WebSocketを破棄】
        }

        //****************************************************************
        //【AllSessionClose】WebSocket接続毎の処理
        //****************************************************************
        public void AllSessionClose()
        {
            Parallel.ForEach(_client, p => //【並列処理をすべて強制終了させる】
            {
                if (p.State == WebSocketState.Open)
                {
                    p.CloseAsync(WebSocketCloseStatus.NormalClosure, "", System.Threading.CancellationToken.None);
                }
            }
            );
        }




        //****************************************************************
        //****************************************************************
        //【WebSocketクライアント】
        //****************************************************************
        //****************************************************************
        public static ClientWebSocket client_ws;
        public static byte[] RxBuffer = new byte[KS_FRAMEWIDTH * KS_FRAMEHEIGHT * 4];  //【受信バッファ】(Client)
        public static byte[] TxBuffer = new byte[KS_FRAMEWIDTH * KS_FRAMEHEIGHT * 4];  //【送信バッファ】(Client)

        //****************************************************************
        //【WebSocketClient】WebSocketサーバーに接続する
        //****************************************************************
        public static async void WebSocketClient()
        {
            client_ws = new ClientWebSocket();
            client_ws.Options.SetBuffer(0x10000, 0x10000); //【バッファサイズは最大65536まで】

            try
            {
                await client_ws.ConnectAsync( //【サーバーに接続要求】
                   new Uri("ws://localhost:8080/"),
                   CancellationToken.None
                );
            }
            catch { }
        }

        //****************************************************************
        //【ClientReceive】WebSocketサーバーからデータを受信する
        //****************************************************************
        public static async void ClientReceive()
        {
            while (true)
            {
                if (client_ws.State == WebSocketState.Open)
                {
                    try
                    {
                        await client_ws.ReceiveAsync(
                           new ArraySegment<byte>(RxBuffer),
                           CancellationToken.None
                        );
                        //【受信後実行したいコードをここに記述する】
                    }
                    catch { }
                }
            }
        }

        //****************************************************************
        //【ClientSendImage】WebSocketサーバーにイメージデータを送信する
        //****************************************************************
        public static async void ClientSendImage()
        {
            while (true)
            {
                if (client_ws.State == WebSocketState.Open)
                {
                    if (flagSend) //【送信準備は整ったか？】
                    {
                        //【送信処理】
                        flagSend = false;
                        for (int i = 0; i < KS_PixelDataARGB.Length; i += 4)
                        {
                            TxBuffer[i] = KS_PixelDataARGB[i + 2];     //【Blue】
                            TxBuffer[i + 1] = KS_PixelDataARGB[i + 1]; //【Green】
                            TxBuffer[i + 2] = KS_PixelDataARGB[i];     //【Red】
                            TxBuffer[i + 3] = 180; //【α】 KS_PixelDataARGB[i + 3];
                        }
                        try
                        {
                            await client_ws.SendAsync(
                               new ArraySegment<byte>(TxBuffer),
                               WebSocketMessageType.Binary,
                               true,
                               CancellationToken.None
                            );
                        }
                        catch { 
                            //【何かエラーが発生した場合】
                        }
                    }
                }
            }
        }

        //****************************************************************
        //【ClientSessionClose】WebSocketサーバーとの接続を終了する
        //****************************************************************
        public static async void ClientSessionClose()
        {
            if ((client_ws != null) && (client_ws.State == WebSocketState.Open))
            {
                await client_ws.CloseAsync(WebSocketCloseStatus.NormalClosure, string.Empty, CancellationToken.None);
            }
        }

    }
}
